/*
 * Copyright (C) 2004-2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/*
 * This file is based in part on vcut.c from vorbis-tools-1.0.1,
 * (c) 2000-2001 Michael Smith <msmith@xiph.org>
 */

#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

#include "server.h"

struct vcut_packet{
        int length;
        unsigned char *packet;
};

struct para_ogg_state {
	ogg_sync_state *sync_in;
	ogg_stream_state *stream_in, *stream_out;
	vorbis_dsp_state *vd;
	vorbis_block *vb;
	vorbis_info *vi;
	unsigned int serial;
	struct vcut_packet **headers; /* 3 */
	FILE *in;
	OggVorbis_File vf;
};

static struct para_ogg_state *s;
static int packetno;
static ogg_int64_t offset_granulepos, granpos;
static char *ogg_buf;
static ssize_t ogg_buf_len;


static struct vcut_packet *save_packet(ogg_packet *packet)
{
        struct vcut_packet *p = para_malloc(sizeof(struct vcut_packet));

	clog(DEBUG, "%s: saving packet (%d bytes)\n", __func__, packet->bytes);
        p->length = packet->bytes;
        p->packet = para_malloc(p->length);
        memcpy(p->packet, packet->packet, p->length);

        return p;
}

static void free_headers(void)
{
	int i;

	clog(DEBUG, "%s: freeing headers\n", __func__);
	for (i = 0; i < 3; i++) {
		free(s->headers[i]->packet);
		free(s->headers[i]);
	}
}

static int page_out(ogg_page *page)
{
	return ogg_stream_flush(s->stream_out, page);
}

static int update_sync()
{
        unsigned char *buffer = ogg_sync_buffer(s->sync_in, 4096);
        int bytes = fread(buffer, 1, 4096, s->in);
        ogg_sync_wrote(s->sync_in, bytes);
        return bytes;
}

/* Pull out and save the 3 header packets from the input file.
 */
static int process_headers(void)
{
        vorbis_comment vc;
        ogg_page page;
        ogg_packet packet;
        int i;

	clog(DEBUG, "%s: \n", __func__);
        ogg_sync_init(s->sync_in);
        vorbis_info_init(s->vi);
        vorbis_comment_init(&vc);
	if (update_sync() < 0)
		return -1;
	if (ogg_sync_pageout(s->sync_in, &page) != 1){
		clog(ERROR, "%s: input not ogg\n", __func__);
		return -1;
	}
        s->serial = ogg_page_serialno(&page);
	ogg_stream_init(s->stream_in, s->serial);
        if (ogg_stream_pagein(s->stream_in, &page) < 0) {
		clog(ERROR, "%s: error in first page\n", __func__);
		return -1;
        }
	if (ogg_stream_packetout(s->stream_in, &packet) != 1) {
		clog(ERROR, "%s: error in first packet\n", __func__);
		return -1;
	}
        if (vorbis_synthesis_headerin(s->vi, &vc, &packet) < 0) {
		clog(ERROR, "%s: error in primary header: not vorbis?\n",
			__func__);
		return -1;
        }
	s->headers[0] = save_packet(&packet);

        i = 0;
        while (i < 2) {
		while (i < 2) {
			int res = ogg_sync_pageout(s->sync_in, &page);
			if (!res)
				break;
			if (res != 1)
				continue;
			ogg_stream_pagein(s->stream_in, &page);
			while (i < 2) {
				res = ogg_stream_packetout(s->stream_in,
					&packet);
				if (!res)	
					break;
				if (res < 0) {
					clog(ERROR, "%s: Secondary "
						"header corrupt\n", __func__);
					return -1;
				}
				s->headers[i + 1] = save_packet(&packet);
				vorbis_synthesis_headerin(s->vi, &vc, &packet);
				i++;
			}
		}
		if (update_sync() <= 0) {
			clog(ERROR, "%s: EOF in headers or header corrupt\n",
				__func__);
			return -1;
		}
	}
        vorbis_comment_clear(&vc);
	clog(DEBUG, "%s: success\n", __func__);
        return 0;
}

static void submit_headers_to_stream(void)
{
        int i;

	clog(DEBUG, "%s: submitting three headers\n", __func__);
	for (i = 0; i < 3; i++) {
		ogg_packet p;
		p.bytes = s->headers[i]->length;
		p.packet = s->headers[i]->packet;
		p.b_o_s = ((i == 0)? 1 : 0);
		p.e_o_s = 0;
		p.granulepos = -1;
		p.packetno = 0;
		ogg_stream_packetin(s->stream_out, &p);
	}
}

static int init_oggvorbis_file(void)
{
	ogg_page p;
	ssize_t len;

	memset(s->headers, 0, sizeof(struct vcut_packet) * 3);
	rewind(s->in);
	if (process_headers() < 0)
		return -1;
	clog(DEBUG, "%s: synthesis_init\n", __func__);
	vorbis_synthesis_init(s->vd, s->vi);
	clog(DEBUG, "%s: block_init\n", __func__);
	vorbis_block_init(s->vd, s->vb);
	clog(DEBUG, "%s: stream_out init\n", __func__);
	ogg_stream_init(s->stream_out, s->serial);
	clog(DEBUG, "%s: submit headers\n");
	submit_headers_to_stream();
	free_headers();
	packetno = 4;
	offset_granulepos = 0; /* nec. */
	granpos = 0; /* nec. */
	ogg_buf_len = 0;
	while (page_out(&p)) {
		len = p.body_len + p.header_len;
		ogg_buf = para_realloc(ogg_buf, ogg_buf_len + len);
		memcpy(ogg_buf + ogg_buf_len, p.header, p.header_len);
		memcpy(ogg_buf + ogg_buf_len + p.header_len, p.body,
			p.body_len);
		ogg_buf_len += len;
	}
	return ogg_buf_len > 0? ogg_buf_len : -1;
}

static int ogg_get_file_info(FILE *file, char *info_str, long unsigned *frames,
	int *seconds)
{
	int ret;
	vorbis_info *vi;
	
	s->in = file;
	if (!file) {
		clog(CRIT, "%s: invalid file pointer\n", __func__);
		return -1;
	}
	rewind(file);
	ret = ov_open(file, &s->vf, NULL, 0);
	if (ret < 0) {
		/*
                 * When you use ov_open() for this, you should fclose()
                 * the file pointer if, and only if, the ov_open() call
                 * fails. If it succeeds, you must call ov_clear() to clear
                 * the decoder's buffers and close the file for you.
		 */
		clog(ERROR, "%s: ogg format not recognized\n", __func__);
		fclose(file);
		return ret;
	}
	vi = ov_info(&s->vf, 0);
	if (!vi)
		return -1;
	*frames = ov_pcm_total(&s->vf, -1);
	*seconds = ov_time_total(&s->vf, -1);
	sprintf(info_str, "audio_file_info1:%ldkHz, %d channels, %ldkbps, "
		"%lu samples\n"
		"audio_file_info2: \n"
		"audio_file_info3: \n",
		vi->rate / 1000, vi->channels, ov_bitrate(&s->vf, 0) / 1000,
		*frames);
	return init_oggvorbis_file();
}

static int get_page(ogg_page *page)
{
	int ret;

	for (;;) {
	//	clog(DEBUG, "%s: sync_pageout\n", __func__);
		ret = ogg_sync_pageout(s->sync_in, page);
		if (ret < 0)
			continue;
		if (!ret) {
			update_sync();
			continue;
		}
//		clog(DEBUG, "%s: page in\n", __func__);
		ret = ogg_stream_pagein(s->stream_in, page);
		if (ret < 0) {
			clog(DEBUG, "%s: stream_pagein failed\n", __func__);
			return -1;
		}
		if (ogg_page_eos(page))
			return 0;
		return 1;
	}
}

static void shuffle_packets(ogg_int64_t granpos)
{
	int ret;
	ogg_packet packet;

	while ((ret = ogg_stream_packetout(s->stream_in, &packet))) {
		if (ret < 0) {
			clog(ERROR, "%s: packetout failed (%d)\n", __func__,
				ret);
			continue;
		}
		packet.packetno = packetno++;
		packet.granulepos = granpos;
		ogg_stream_packetin(s->stream_out, &packet);
	}
}

static void ogg_handle_repos_request(long unsigned *request)
{
	int ret;
	ogg_page page;

	rewind(s->in);
	ret = ov_pcm_seek_page(&s->vf, *request);
	clog(DEBUG, "%s: ov_seek_page returned %d\n", __func__, ret);
again:
	ret = get_page(&page);
	if (ret < 0) {
		clog(ERROR, "%s: get_page failed\n", __func__);
		return;
	}
	offset_granulepos = ogg_page_granulepos(&page);
	clog(DEBUG, "%s: current offset: %d\n", __func__, offset_granulepos);
	if (offset_granulepos < *request)
		goto again;
	clog(DEBUG, "%s: using offset %d\n", __func__, offset_granulepos);
	*request = offset_granulepos;
	shuffle_packets(granpos);
	page_out(&page);
}

static void ogg_end_of_input(FILE **sw, pid_t *pid)
{
	clog(DEBUG, "%s: ogg_sync_destroy\n", __func__);
	ogg_sync_destroy(s->sync_in);
	clog(DEBUG, "%s: ogg_stream_destroy\n", __func__);
	ogg_stream_destroy(s->stream_in);
	ogg_stream_destroy(s->stream_out);
	s->stream_out = para_malloc(sizeof(ogg_stream_state));
	s->stream_in = para_malloc(sizeof(ogg_stream_state));
	s->sync_in = para_malloc(sizeof(ogg_sync_state));
	if (s->in) {
		clog(DEBUG, "%s: ov_clear\n", __func__);
		ov_clear(&s->vf);
	}
	clog(DEBUG, "%s: vorbis_dsp_clear\n", __func__);
	vorbis_dsp_clear(s->vd);
	clog(DEBUG, "%s: vorbis_block_clear\n", __func__);
	vorbis_block_clear(s->vb);
	clog(DEBUG, "%s: vorbis_info_clear\n", __func__);
	vorbis_info_clear(s->vi);
#if 1
	if (*sw && *pid) {
		clog(DEBUG, "%s: closing stream writer (fd %d)\n", __func__,
			fileno(*sw));
		pclose(*sw);
		kill(*pid, SIGTERM);
		/* FIXME This blocks server connections */
		clog(DEBUG, "%s: waiting for stream writer (pid %d)\n", 
			__func__, *pid);
		waitpid(*pid, NULL, 0);
		*pid = 0;
	}
	*sw = NULL;
#endif
	//usleep(5000 * 1000);

	if (ogg_buf) {
		/* last write was a short write */
		free(ogg_buf);
		ogg_buf = NULL;
	}
}

static int ogg_send_chunk(FILE **file, char *info_str, long unsigned *num_pcm,
	int afs_status, pid_t *pid)
{
	int ret;
	ogg_page p;
	ogg_int64_t tmp;
	
	if (packetno < 5)
		sleep(1); 	/* give the clients the chance to connect 
				 * before streaming starts. Crap.
				 */
	if (ogg_buf)
		goto write;
	
	if (afs_status & AFS_REPOS)
		ogg_handle_repos_request(num_pcm);
	if (afs_status & AFS_NEXT) {
		ogg_end_of_input(file, pid);
		return 0;
	}
get_page:
	ret = get_page(&p);
	if (ret <= 0)
		goto out;
	tmp = ogg_page_granulepos(&p);
	if (tmp >= offset_granulepos)
		granpos = tmp - offset_granulepos;
	shuffle_packets(granpos);
	if (!page_out(&p)) {
		clog(DEBUG, "%s: pageout failed\n", __func__);
		goto get_page;
	}
	if (!ogg_buf) {
		ogg_buf_len = p.body_len + p.header_len;
		ogg_buf = para_malloc((ogg_buf_len + 1) * sizeof(char));
		memcpy(ogg_buf, p.header, p.header_len);
		memcpy(ogg_buf + p.header_len, p.body, p.body_len);
	}
write:
	ret = write(fileno(*file), ogg_buf, ogg_buf_len);
	if (ret < 0)
		clog(ERROR, "%s: can not write to fd %d (%s)\n", __func__,
			fileno(*file), strerror(errno));
//	clog(DEBUG, "%s: wrote: %d/%d\n", __func__, ret, len);
	if (ret <= 0)
		goto out;
	*num_pcm = granpos + offset_granulepos;
	if (ret < ogg_buf_len) {
		clog(DEBUG, "%s: short write (%i/%i)\n", __func__, ret,
			ogg_buf_len);
		memcpy(ogg_buf, ogg_buf + ret, ogg_buf_len - ret);
		ogg_buf_len -= ret;
		return ret;
	}
	free(ogg_buf);
	ogg_buf = NULL;
	if (afs_status & AFS_NEXT) {
		ogg_end_of_input(file, pid);
		return 0;
	}
	return ret;
out:	
	ogg_end_of_input(file, pid);
	return ret;
}


int ogg_init(void *p)
{
	struct audio_format *af = p;
	s = para_malloc(sizeof(struct para_ogg_state));

	af->get_file_info = ogg_get_file_info,
	af->send_chunk = ogg_send_chunk;
	af->open_cmd = "pogg-http -q -";

	memset(s, 0, sizeof(struct para_ogg_state));
	s->sync_in = para_malloc(sizeof(ogg_sync_state));
	s->stream_in = para_malloc(sizeof(ogg_stream_state));
	s->stream_out = para_malloc(sizeof(ogg_stream_state));
	s->vd = para_malloc(sizeof(vorbis_dsp_state));
	s->vi = para_malloc(sizeof(vorbis_info));
	s->vb = para_malloc(sizeof(vorbis_block));
	s->headers = para_malloc(sizeof(struct vcut_packet) * 3);
	return 1;
}
